(***********************************************************************)
(*                                                                     *)
(*                               Grimage                               *)
(*                                                                     *)
(*                           Graphics image                            *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2014-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: colormapping.ml,v 1.7 2014/02/27 16:45:22 fclement Exp $ *)

(* Colormappings. *)

open Colors;;

(* Data type definitions. *)

type color_scale = Colors.t array;;
type from_float = float and to_float = float;;
type mapping = color_scale -> from_float -> to_float -> float -> Colors.t;;

type num_steps = int;;
type from_color = Colors.t and to_color = Colors.t;;
type color_scaler = from_color -> to_color -> num_steps -> int -> Colors.t;;

(* Scalers. *)

(*
 Scaling integer range [n..m] using [w] steps. Any integer [i] in the range
 [0..w] is mapped to the corresponding value into the range [n..m], since we
 have: [scale n m w i = (m * i + n * (w - i)) / w]

 The function extends to the full range of integers by clipping: its argument
 [i] is automatically clipped into the range [0..w]. Hence, if [i] is negative
 its scale is [n] and if [i >= w] its scale is [m]).
*)
let scale n m w i =
  assert (w > 0);
  if i <= 0 then n else
  if i >= w then m else
  let mid = (w + 1) / 2 in
  let round_quot w x = (x + mid) / w in
  round_quot w (m * i + n * (w - i))
;;

(* Scaling a color from [fc] to [tc], by [i] steps over [w] steps. *)
let make_scaler from_color to_color fc tc w i =
  let (fx, fy, fz) = from_color fc
  and (tx, ty, tz) = from_color tc in
  to_color (scale fx tx w i) (scale fy ty w i) (scale fz tz w i)
;;

let scaler_hsb = make_scaler Colors.hsb_of_color Colors.color_of_hsb;;

(* For efficiency reasons, we do not use the [make_scaler] generator to
   scale RGB colors, since we do not want to build 2 triples at each
   scaler_rgb call. *)
(*let scaler_rgb = make_scaler rgb_of_color color_of_rgb;;*)
let scaler_rgb fc tc w i =
  Colors.rgb
    (scale (r_of fc) (r_of tc) w i)
    (scale (g_of fc) (g_of tc) w i)
    (scale (b_of fc) (b_of tc) w i)
;; 

(* Mappings. *)

let round_pos x = int_of_float (x +. 0.5)
and round_neg x = int_of_float (x -. 0.5)
;;
let round x = if x >= 0. then round_pos x else round_neg x;;

let linear_mapping cscale x1 x2 x =
  let w = Array.length cscale in
  let xr = (x -. x1) /. (x2 -. x1) in
  let i = round (float_of_int w *. xr) in
  (* May be we can do
     let k = float_of_int w /. (x2 -. x1) in
     fun x ->
       let i = round (w * (x -. x1)) in
     
     More precisely we should clip x into the range [x1..x2]
     before computing i? *)
  (* We clip within the range [0..w-1]. *)
  (* let i = max 0 (min i (w - 1)) in *)
  let i =
    if i <= 0 then 0 else
    if i >= w then w - 1 else
    i in
  cscale.(i)
;;

(* Colormappings. *)

let make_color_scale scaler fc tc w = Array.init w (scaler fc tc w);;

type t =
  {
   color_scale : color_scale;
   from_float : from_float;
   to_float : to_float;
   mapping : mapping;
  }
;;

let make m cs ff tf =
  {
   color_scale = cs;
   from_float = ff;
   to_float = tf;
   mapping = m;
  }
;;

let make_linear scaler fc tc w =
  make linear_mapping (make_color_scale scaler fc tc w)
;;

let make_rgb = make_linear scaler_rgb
and make_hsb = make_linear scaler_hsb
;;

let get_color cm = cm.mapping cm.color_scale cm.from_float cm.to_float;;
